/*
 * Copyright (c) 2008-2010 Lu, Chao-Ming (Tetralet).  All rights reserved.
 * 
 * This file is part of LilyTerm.
 *
 * LilyTerm is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * LilyTerm is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with LilyTerm.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "menu.h"

// Sometimes we may popup a menu on a non-active (not focus) window...
extern GtkWidget *menu_active_window;
extern struct Page_Color page_color[PAGE_COLOR];
extern gchar *profile_dir;

void create_menu(struct Window *win_data)
{
#ifdef DETAIL
	g_debug("! Launch create_menu() for win_data %p", win_data);
#endif
	GtkWidget *menu_item, *sub_menu=NULL;
	gsize i=0;
	//gboolean enable_vte_cjk_width = win_data->VTE_CJK_WIDTH ? 1 : 0;
	//g_debug("enable_vte_cjk_width = %d, and VTE_CJK_WIDTH (environ) = %s",
	//	enable_vte_cjk_width, g_getenv("VTE_CJK_WIDTH"));
	win_data->menu = gtk_menu_new();

	gboolean have_encoding_locale_menu = 0;
	if (win_data->supported_encodings)
	{
		gint count_menu=0;
		GtkWidget *sub_menu_item = create_sub_item_subitem (_("Change text encoding"), GTK_STOCK_DND);
		win_data->encoding_sub_menu = gtk_menu_new ();
		gtk_menu_item_set_submenu (GTK_MENU_ITEM (sub_menu_item), win_data->encoding_sub_menu);
		
		gchar **lists = get_supported_locales ((gchar *)vte_terminal_get_encoding(
								VTE_TERMINAL(win_data->current_vte)),
							win_data->encodings_list,
							(gchar *) get_encoding_from_locale(),
							NULL);
		
		while (lists[i]!=NULL)
		{
			if (lists[i][0] == '\0')
			{
				i++;
				continue;
			}
			// if NULL: Not find in sub_menu
			if (!(check_name_in_menuitem(win_data->encoding_sub_menu, lists[i])))
			{
				add_menuitem_to_encoding_sub_menu(win_data, i, lists[i]);
				count_menu ++;
			}
			i++;
		}
		if (count_menu>1)
		{
			gtk_menu_shell_append (GTK_MENU_SHELL (win_data->menu), sub_menu_item);
			have_encoding_locale_menu = 1;
		}
		g_strfreev(lists);
	}

	if (win_data->supported_locales)
	{
		int j;

		for (j=0; j<2; j++)
		{
			gint count_menu=0;
			GtkWidget *sub_menu_item;
			if (j)
				sub_menu_item = create_sub_item_subitem (_("New tab"), GTK_STOCK_DND_MULTIPLE);
			else
				sub_menu_item = create_sub_item_subitem (_("New tab with specified locale"),
									 GTK_STOCK_DND_MULTIPLE);
			win_data->locale_sub_menu = gtk_menu_new ();
			gtk_menu_item_set_submenu (GTK_MENU_ITEM (sub_menu_item), win_data->locale_sub_menu);
	
			// get all locales list
			gchar **lists = get_supported_locales ( win_data->environ_locale,
								win_data->locales_list,
								get_default_LC_TYPE(),
								NULL);
			i=0;
			while (lists[i]!=NULL)
			{
				if ((lists[i][0] == '\0') || ((j==0) && (g_ascii_strcasecmp (lists[i], "UTF-8")==0)))
				{
					i++;
					continue;
				}
				// if NULL: Not find in sub_menu
				if (!(check_name_in_menuitem(win_data->locale_sub_menu, lists[i])))
					count_menu += add_menuitem_to_locale_sub_menu(win_data, i, lists[i]);
				i++;
			}
			if (count_menu>1)
			{
				// if the total of menu item is >1, add it to menu.
				gtk_menu_shell_append (GTK_MENU_SHELL (win_data->menu), sub_menu_item);
				have_encoding_locale_menu = 1;
				// don't create "New tab" submenu.
				j=2;
			}
			g_strfreev(lists);
		}
	}

	win_data->menuitem_profile_list = create_sub_item_subitem (_("New window with specified profile"), GTK_STOCK_NEW);
	win_data->subitem_profile_list = create_sub_item_submenu (win_data->menu, win_data->menuitem_profile_list);
	create_profile_list(win_data->subitem_profile_list, GTK_STOCK_NEW, win_data);
	win_data->profile_dir_modtime = get_profile_dir_modtime();
	// g_debug("win_data->profile_dir_modtime = %ld", win_data->profile_dir_modtime);

	if (have_encoding_locale_menu)
	{
		menu_item = gtk_separator_menu_item_new ();
		gtk_menu_shell_append (GTK_MENU_SHELL (win_data->menu), menu_item);
	}

//	// ----------------------------------------
//	menu_item = gtk_separator_menu_item_new ();
//	gtk_menu_shell_append (GTK_MENU_SHELL (win_data->menu), menu_item);

	// Change the font for every tab
	create_menu_item (0, win_data->menu, _("Change the font"), NULL, GTK_STOCK_SELECT_FONT,
			  (GSourceFunc)select_font, win_data->window);

#ifdef ENABLE_GDKCOLOR_TO_STRING
	if (win_data->show_color_selection_menu)
	{
		// The submenu of Change color
		sub_menu = create_sub_item (win_data->menu, _("Change colors"), GTK_STOCK_SELECT_COLOR);

		// Change the foreground color for every tab
		create_menu_item (0, sub_menu, _("Change the foreground color"), NULL, GTK_STOCK_SELECT_COLOR,
				  (GSourceFunc)dialog, GINT_TO_POINTER (9));

		// Change the background color for every tab
		create_menu_item (0, sub_menu, _("Change the background color"), NULL, GTK_STOCK_SELECT_COLOR,
				  (GSourceFunc)dialog, GINT_TO_POINTER (10));

		// Change the background color for every tab
		create_menu_item (0, sub_menu, _("Adjust the brightness of ANSI colors"), NULL, GTK_STOCK_SELECT_COLOR,
				  (GSourceFunc)dialog, GINT_TO_POINTER (19));

		if (win_data->use_color_page)
		{
			// ----------------------------------------
			add_separator_menu (sub_menu);

			for (i=0; i<PAGE_COLOR; i++)
			{
				gchar *label_name = g_strdup_printf(_("Change the %s color on tab"),
								    page_color[i].comment);
				create_menu_item (0, sub_menu, label_name, NULL, GTK_STOCK_SELECT_COLOR,
						  (GSourceFunc)dialog, GINT_TO_POINTER (11+i));
				g_free(label_name);
			}
		}
	}
	// ----------------------------------------
	add_separator_menu (win_data->menu);

#endif
	if (win_data->show_transparent_menu)
	{
#ifdef ENABLE_RGBA
		if (win_data->use_rgba)
		{
			// Transparent Window
			win_data->menuitem_trans_win = create_menu_item (1, win_data->menu, _("Transparent window"),
									 NULL, NULL, (GSourceFunc)set_trans_win,
									 win_data->window);
		
			// Window opacity
			create_menu_item (0, win_data->menu, _("Window opacity"), NULL, GTK_STOCK_EXECUTE,
					  (GSourceFunc)dialog, GINT_TO_POINTER (8));
		
			// ----------------------------------------
			add_separator_menu (win_data->menu);
		}
#endif
		// Transparent Background
		win_data->menuitem_trans_bg = create_menu_item (1, win_data->menu, _("Transparent background"), NULL, NULL,
								(GSourceFunc)set_trans_bg, win_data);
	
		create_menu_item (0, win_data->menu, _("Background saturation"), NULL, GTK_STOCK_EXECUTE,
				  (GSourceFunc)dialog, GINT_TO_POINTER (2));

		// ----------------------------------------
		add_separator_menu (win_data->menu);
	}

	if (win_data->show_copy_paste_menu)
	{
		if (win_data->enable_hyperlink)
			// copy_url
			win_data->menuitem_copy_url = create_menu_item (0, win_data->menu, _("Copy URL"), NULL, 
									GTK_STOCK_COPY, (GSourceFunc)copy_url_clipboard,
									NULL);

		// copy
		win_data->menuitem_copy = create_menu_item (0, win_data->menu, _("Copy"), NULL, GTK_STOCK_COPY, 
							    (GSourceFunc)copy_clipboard, NULL);

		// paste
		win_data->menuitem_paste = create_menu_item (0, win_data->menu, _("Paste"), NULL, GTK_STOCK_COPY,
							     (GSourceFunc)paste_clipboard, NULL);

		// ----------------------------------------
		add_separator_menu (win_data->menu);
	}

	// use scrollback lines
	// g_debug("menuitem_hide_scroll_bar)->active = %d", win_data->scrollback_lines ? 1 : 0);
	// GTK_CHECK_MENU_ITEM(menuitem_hide_scroll_bar)->active = win_data->scrollback_lines ? 1 : 0;
	win_data->menuitem_hide_scroll_bar = create_menu_item (1, win_data->menu, _("Scrollback lines"), NULL, NULL,
							       (GSourceFunc)hide_scrollback_lines, NULL);
	GTK_CHECK_MENU_ITEM(win_data->menuitem_hide_scroll_bar)->active = win_data->show_scrollbar;

	// clean scrollback lines
	create_menu_item (0, win_data->menu, _("Clean scrollback lines"), NULL, GTK_STOCK_CLEAR,
			  (GSourceFunc)clean_scrollback_lines, NULL);

	// ----------------------------------------
	add_separator_menu (win_data->menu);

	if (win_data->dim_text)
		// Dim when inactive
	       	win_data->menuitem_dim_text = create_menu_item (1, win_data->menu, _("Dim text when inactive"), NULL, NULL,
								(GSourceFunc)set_dim_text, win_data);

	// Cursor Blinks
	win_data->menuitem_cursor_blinks = create_menu_item (1, win_data->menu, _("Cursor blinks"), NULL, NULL,
								(GSourceFunc)set_cursor_blinks, win_data);

	// Audible bell
	win_data->menuitem_audible_bell = create_menu_item (1, win_data->menu, _("Audible bell"), NULL, NULL,
								(GSourceFunc)set_audible_bell, win_data);

	// Show the tabs bar
	win_data->menuitem_show_tabs_bar = create_menu_item (0, win_data->menu, _("Show the tabs bar"), NULL,
							     GTK_STOCK_GOTO_BOTTOM,
							     (GSourceFunc)launch_hide_and_show_tabs_bar,
							     GINT_TO_POINTER(TRUE));
	// Hide the tabs bar
	win_data->menuitem_hide_tabs_bar = create_menu_item (0, win_data->menu, _("Hide the tabs bar"), NULL,
							     GTK_STOCK_GOTO_TOP,
							     (GSourceFunc)launch_hide_and_show_tabs_bar,
							     GINT_TO_POINTER(FALSE));

	// ----------------------------------------
	add_separator_menu (win_data->menu);


	// Input Method
	if (win_data->show_input_method_menu)
	{
		sub_menu = create_sub_item (win_data->menu, _("Switch input methods"), GTK_STOCK_INDEX);
		vte_terminal_im_append_menuitems (VTE_TERMINAL(win_data->current_vte), GTK_MENU_SHELL (sub_menu));
	}

	// Edit tab name
	if (win_data->show_change_page_name_menu)
		create_menu_item (0, win_data->menu, _("Rename this tab"), NULL, GTK_STOCK_EDIT,
			  (GSourceFunc)dialog, GINT_TO_POINTER(1));

	if (win_data->show_input_method_menu || win_data->show_change_page_name_menu)
		add_separator_menu (win_data->menu);

	if (win_data->show_resize_menu)
	{
		// Window Size
		create_menu_item (0, win_data->menu, _("Increase window size"), NULL, GTK_STOCK_ZOOM_IN,
				  (GSourceFunc)set_vte_font, GINT_TO_POINTER(3));

		create_menu_item (0, win_data->menu, _("Decrease window size"), NULL, GTK_STOCK_ZOOM_OUT,
				  (GSourceFunc)set_vte_font, GINT_TO_POINTER(4));

		// ----------------------------------------
		add_separator_menu (win_data->menu);

		// Reset font and window size
		create_menu_item (0, win_data->menu, _("Reset to user custom font/size"), NULL, GTK_STOCK_ZOOM_100,
				  (GSourceFunc)set_vte_font, GINT_TO_POINTER(5));
		
		create_menu_item (0, win_data->menu, _("Reset to system font/size"), NULL, GTK_STOCK_ZOOM_100,
				  (GSourceFunc)set_vte_font, GINT_TO_POINTER(6));
	}
	
	// Reset
	create_menu_item (0, win_data->menu, _("Reset the content"), NULL, GTK_STOCK_REFRESH,
			  (GSourceFunc)reset_vte, NULL);

	// ----------------------------------------
	add_separator_menu (win_data->menu);

	// Usage
	create_menu_item (0, win_data->menu, _("Usage"), NULL, GTK_STOCK_HELP,
				  (GSourceFunc)dialog, GINT_TO_POINTER(5));

	// Get the key value for using in profile
	create_menu_item (0, win_data->menu, _("Get function key value"), NULL, GTK_STOCK_CONVERT,
				  (GSourceFunc)dialog, GINT_TO_POINTER(4));

	// The submenu of Change color
	sub_menu = create_sub_item (win_data->menu, _("User profile"), GTK_STOCK_PROPERTIES);

	// Profile
	//menu_item = gtk_image_menu_item_new_with_label(_("Profile sample"));
	create_menu_item (0, sub_menu, _("Save settings"), NULL, GTK_STOCK_SAVE,
			  (GSourceFunc)save_user_settings, NULL);

	create_menu_item (0, sub_menu, _("Save settings as..."), NULL, GTK_STOCK_SAVE_AS,
			  (GSourceFunc)save_user_settings_as, NULL);

//	// ----------------------------------------
//	menu_item = gtk_separator_menu_item_new ();
//	gtk_menu_shell_append (GTK_MENU_SHELL (sub_menu), menu_item);
//
//	menu_item = gtk_image_menu_item_new_with_label(_("Load settings from..."));
//	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menu_item),
//				      gtk_image_new_from_stock(GTK_STOCK_SAVE, GTK_ICON_SIZE_MENU));
//	gtk_menu_shell_append(GTK_MENU_SHELL(sub_menu), menu_item);
//	g_signal_connect(menu_item, "activate", G_CALLBACK(load_user_settings_from_dialog), win_data);

	if (win_data->show_exit_menu)
	{
		// Exit LilyTerm
		gchar *exit_str = g_strdup_printf(_("Quit %s"), PACKAGE_NAME);
		create_menu_item (0, win_data->menu, exit_str, NULL, GTK_STOCK_QUIT,
				  (GSourceFunc)main_quit, win_data);
		g_free(exit_str);
	}

	gtk_widget_show_all(win_data->menu);
}


void clean_scrollback_lines(GtkWidget *widget, gpointer user_data)
{
#ifdef DETAIL
	g_debug("! Launch clean_scrollback_lines()!");
#endif
	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(menu_active_window), "Win_Data");
	vte_terminal_set_scrollback_lines (VTE_TERMINAL(win_data->current_vte), 0);
	vte_terminal_set_scrollback_lines (VTE_TERMINAL(win_data->current_vte), win_data->scrollback_lines);
}

void copy_url_clipboard(GtkWidget *widget, gpointer user_data)
{
#ifdef DETAIL
	g_debug("! Launch copy_url_clipboard()!");
#endif
	if (gtk_widget_get_name(widget))
	{
		extern GtkClipboard *pclipboard;
		gtk_clipboard_set_text(pclipboard, gtk_widget_get_name(widget), -1);
	}
}

void copy_clipboard(GtkWidget *widget, gpointer user_data)
{
#ifdef DETAIL
	g_debug("! Launch copy_clipboard()!");
#endif
	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(menu_active_window),
								     "Win_Data");
	vte_terminal_copy_clipboard(VTE_TERMINAL(win_data->current_vte));
}

void paste_clipboard(GtkWidget *widget, gpointer user_data)
{
#ifdef DETAIL
	g_debug("! Launch paste_clipboard()!");
#endif
	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(menu_active_window),
								     "Win_Data");
	vte_terminal_paste_clipboard(VTE_TERMINAL(win_data->current_vte));
}

void set_dim_text(GtkWidget *menuitem_dim_text, struct Window *win_data)
{
#ifdef DETAIL
	g_debug("! Launch set_dim_text() with win_data = %p", win_data);
#endif
	win_data->dim_text = GTK_CHECK_MENU_ITEM(menuitem_dim_text)->active;
	dim_vte_text(win_data, NULL, 2);
}

void set_cursor_blinks(GtkWidget *menuitem_cursor_blinks, struct Window *win_data)
{
#ifdef DETAIL
	g_debug("! Launch set_cursor_blinks() with win_data = %p", win_data);
#endif
	win_data->cursor_blinks = GTK_CHECK_MENU_ITEM(menuitem_cursor_blinks)->active;

	gint i;
	GtkWidget *vte;

	for (i=0;i<gtk_notebook_get_n_pages(GTK_NOTEBOOK(win_data->notebook));i++)
	{
		vte=(GtkWidget *)g_object_get_data(G_OBJECT(gtk_notebook_get_tab_label(GTK_NOTEBOOK(
								    win_data->notebook),
									gtk_notebook_get_nth_page(
									    GTK_NOTEBOOK(win_data->notebook),
									    i))),
						   "VteBox");
		vte_terminal_set_cursor_blinks (VTE_TERMINAL(vte), win_data->cursor_blinks);
	}
}

void set_audible_bell(GtkWidget *menuitem_audible_bell, struct Window *win_data)
{
#ifdef DETAIL
	g_debug("! Launch set_audible_bell() with win_data = %p", win_data);
#endif
	win_data->audible_bell = GTK_CHECK_MENU_ITEM(menuitem_audible_bell)->active;
	gint i;
	GtkWidget *vte;

	for (i=0;i<gtk_notebook_get_n_pages(GTK_NOTEBOOK(win_data->notebook));i++)
	{
		vte=(GtkWidget *)g_object_get_data(G_OBJECT(gtk_notebook_get_tab_label(GTK_NOTEBOOK(
								    win_data->notebook),
									gtk_notebook_get_nth_page(
									    GTK_NOTEBOOK(win_data->notebook),
									    i))),
						   "VteBox");
		vte_terminal_set_audible_bell (VTE_TERMINAL(vte), win_data->audible_bell);
	}
}

void launch_hide_and_show_tabs_bar(GtkWidget *widget, gboolean show_tabs_bar)
{
#ifdef DETAIL
	g_debug("! Launch launch_hide_and_show_tabs_bar() with show_tabs_bar = %d", show_tabs_bar);
#endif
	// g_debug("menu_active_window = %p", menu_active_window);

	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(menu_active_window), "Win_Data");
	hide_and_show_tabs_bar(win_data , show_tabs_bar);
}


GtkWidget *check_name_in_menuitem(GtkWidget *sub_menu, const gchar *name)
{
#ifdef DETAIL
	g_debug("! Launch check_name_in_menuitem() with name = %s", name);
#endif
	// FIXME: if sub_menu == NULL...
	GList *widget_list = GTK_MENU_SHELL(sub_menu)->children;
	GtkWidget *return_menuitem = NULL;
	GtkWidget *menu_item = NULL;

	while (widget_list!=NULL)
	{
		menu_item = widget_list->data;
		// g_debug("Got the menu_item = %p", menu_item);
		if (menu_item==NULL) break;

		// g_debug("Checking %s for %s...", gtk_widget_get_name(menu_item), name);
		if (g_ascii_strcasecmp (gtk_widget_get_name(menu_item), name)==0)
			return_menuitem = menu_item;

		if (return_menuitem) return return_menuitem;
		widget_list = widget_list->next;
	}
	return NULL;
}

GtkWidget *add_menuitem_to_encoding_sub_menu(struct Window *win_data,
					     gint no,
					     const gchar *name)
{
#ifdef DETAIL
	g_debug("! Launch add_menuitem_to_encoding_sub_menu() with win_data = %p, no = %d, name = %s",
		win_data, no, name);
#endif
	GtkWidget *menu_item = gtk_radio_menu_item_new_with_label(win_data->encoding_group, name);
	if (no==0)
	{
		win_data->default_encoding_widget = menu_item;
		// win_data->default_encoding_str = g_strdup(name);
	}
	gtk_widget_set_name(menu_item, name);
	// g_debug("Set the %d encoding to %s", no, name);
	gtk_menu_shell_append(GTK_MENU_SHELL(win_data->encoding_sub_menu), menu_item);
	// g_debug("win_data->encoding_group = %p", win_data->encoding_group);
	win_data->encoding_group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (menu_item));
	
	g_signal_connect(menu_item, "activate", G_CALLBACK(set_encoding), NULL);

	return menu_item;
}

gint add_menuitem_to_locale_sub_menu(struct Window *win_data,
					   gint no,
					   gchar *name)
{
#ifdef DETAIL
	g_debug("! Launch add_menuitem_to_locale_sub_menu() with win_data = %p, no = %d, name = %s", 
		win_data, no, name);
#endif
	// if name = "zh_TW.UTF-8" or just "UTF-8"
	gchar **locales = g_strsplit(name, ".", 0);
	gint creat_item = 1;

	if (locales[1])
	{
		if (g_ascii_strcasecmp(locales[1], "UTF-8")==0)
			creat_item = 2;
	}
	else if (g_ascii_strcasecmp(name, "UTF-8")==0)
		creat_item = 2;

	gsize i;
	for (i=0; i<creat_item; i++)
	{
		gchar* item_name;
		if (i)
			item_name = g_strdup_printf("%s%s", name, _(" (Wide)"));
		else
			item_name = name;
		GtkWidget *menu_item = gtk_image_menu_item_new_with_label(item_name);
		gtk_widget_set_name(menu_item, name);
		// g_debug("Set the %d locale to %s", no, item_name);
		gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menu_item),
					      gtk_image_new_from_stock(GTK_STOCK_DND_MULTIPLE, GTK_ICON_SIZE_MENU));
		gtk_menu_shell_append(GTK_MENU_SHELL(win_data->locale_sub_menu), menu_item);
		// g_debug("win_data->locale_group = %p", win_data->locale_group);
		// win_data->locale_group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (menu_item));
		g_signal_connect(menu_item, "activate", G_CALLBACK(new_tab_with_locale), (gsize *)i);
		if (i)
			g_free(item_name);
	}

	g_strfreev(locales);
	return creat_item;
}


//gboolean check_locale_in_menuitem(GList *locale_list, const gchar *name)
//{
//	//g_debug("Calling check_locale_in_menuitem with locale_list = %p, name =%s, total = %d",
//	//	locale_list, name, g_list_length(locale_list));
//	GList *widget_list = NULL;
//	int i;
//
//	for (i=0; i< g_list_length(locale_list); i++)
//	{
//		widget_list = g_list_nth(locale_list, i);
//		// g_debug("Checking the locale %s with %s", (gchar *)widget_list->data, name);
//		if (g_ascii_strcasecmp (widget_list->data, name)==0)
//			return TRUE;
//	}
//	return FALSE;
//}

void reset_vte(GtkWidget *widget, gpointer user_data)
{
#ifdef DETAIL
	g_debug("! Launch reset_vte()!");
#endif
	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(menu_active_window), "Win_Data");
	vte_terminal_reset(VTE_TERMINAL(win_data->current_vte), TRUE, FALSE);
}

void set_trans_bg(GtkWidget *menuitem_trans_bg, struct Window *win_data)
{
#ifdef DETAIL
	g_debug("! Launch set_trans_bg() for win_data %p", win_data);
#endif
	win_data->transparent_background = GTK_CHECK_MENU_ITEM(menuitem_trans_bg)->active;

	gint i;
	GtkWidget *vte;
	
	for (i=0;i<gtk_notebook_get_n_pages(GTK_NOTEBOOK(win_data->notebook));i++)
	{
		vte=(GtkWidget *)g_object_get_data(G_OBJECT(gtk_notebook_get_tab_label(GTK_NOTEBOOK(
								    win_data->notebook),
									gtk_notebook_get_nth_page(
									    GTK_NOTEBOOK(win_data->notebook),
									    i))),
						   "VteBox");
		set_background_saturation (NULL, 0, win_data->background_saturation, vte);
	}
}
#ifdef ENABLE_RGBA
void set_trans_win(GtkWidget *widget, GtkWidget *window)
{
#ifdef DETAIL
	g_debug("! Launch set_trans_win() for window %p!", window);
#endif

	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(window), "Win_Data");
	// g_debug("Get win_data = %d when set_trans_win!", win_data);

	win_data->transparent_window = GTK_CHECK_MENU_ITEM(win_data->menuitem_trans_win)->active;
	set_window_opacity (NULL, 0, win_data->window_opacity, window);
}
#endif

// it is OK to use either zh_TW.Big5 or Big5 here
void set_encoding(GtkWidget *menuitem, GtkWidget *vte)
{
#ifdef DETAIL
	g_debug("! Launch set_encoding() for vte %p", vte);
#endif
	// g_debug("menu_active_window = %p", menu_active_window);
	if (menu_active_window==NULL) return;

	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(menu_active_window), "Win_Data");
	// No action when just right click the mouse
	if (win_data->query_encoding) return;

	const gchar *encoding;
//	// g_debug("Got the encoding item: %p", menuitem);
//	if (menuitem==NULL && vte==NULL)
//	{
//		gchar *err_msg = g_strdup_printf("set_encoding(): menuitem = NULL and vte = NULL\n\n"
//						 "Please report bug to %s, Thanks!",
//						 PACKAGE_BUGREPORT);
//		error_dialog(win_data, err_msg, NULL, 18);
//#ifdef DETAIL
//		g_debug("* free err_msg %p (%s) in set_encoding()", err_msg, err_msg);
//#endif
//		g_free(err_msg);
//		return;
//	}

	// vte==NULL: called by the "activate" signal of menuitem. vte = win_data->current_vte.
	if (vte==NULL)
	{
		// GTK_CHECK_MENU_ITEM(menuitem)->active = false: We only set the encoding with actived menuitem. 
		if (!GTK_CHECK_MENU_ITEM(menuitem)->active) return;
		vte = win_data->current_vte;
	}

	struct Page *page_data = (struct Page *)g_object_get_data(G_OBJECT(vte), "Page_Data");

	if (menuitem)
		encoding = gtk_widget_get_name(menuitem);
	else
	{
		// g_debug("win_data->default_locale = %s", win_data->default_locale);
		gchar **locales = g_strsplit(win_data->environ_locale, ".", -1);
		if (locales[1])
			encoding = win_data->environ_locale;
		else
			encoding = vte_terminal_get_encoding(VTE_TERMINAL(vte));
		g_strfreev(locales);
	}
	g_free (page_data->encoding_str);
	page_data->encoding_str = g_strdup(encoding);
	// g_debug("Trying to set %d page (%p) to encoding %s...", page_data->page_no, vte, encoding);
	
	gchar **locales = g_strsplit(encoding, ".", -1);
	// encoding_str can't be free!
	gchar *encoding_str = locales[1]? locales[1]: locales[0];
	// g_debug("Trying to set %d page (%p) to encoding %s...", page_data->page_no, vte, encoding_str);
	vte_terminal_set_encoding(VTE_TERMINAL(vte), encoding_str);

	// page_data->page_name==NULL: add_page() will call set_encoding() before page_data->label is created.
	// And it will cause the following error: "Document ended unexpectedly while inside an attribute value"
	if (win_data->page_shows_encoding && page_data->page_name!=NULL)
		update_page_name (page_data->window, vte, page_data->page_name, page_data->label_text, page_data->page_no+1,
				  page_data->custom_page_name, page_data->tab_color,
				  page_data->is_root, page_data->is_bold,
				  compare_win_page_encoding(win_data->default_encoding_widget,
				  			    page_data->encoding_str),
				  page_data->encoding_str, page_data->custom_window_title,
				  FALSE);
	g_strfreev(locales);
}

void new_tab_with_locale(GtkWidget *local_menuitem, gboolean VTE_CJK_WIDTH)
{
#ifdef DETAIL
	g_debug("! Launch new_tab_with_locale() with locale = %s and VTE_CJK_WIDTH = %d",
		gtk_widget_get_name(local_menuitem), VTE_CJK_WIDTH);
#endif
	// g_debug("local_menuitem->name = %s", gtk_widget_get_name(local_menuitem);
	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(menu_active_window), "Win_Data");
	// g_debug("Got the VTE_CJK_WIDTH = %s", gtk_widget_get_name(local_menuitem);
	struct Page *page_data = (struct Page *)g_object_get_data(G_OBJECT(win_data->current_vte), "Page_Data");

	// gchar *encoding = get_encoding_from_menu_item_name(local_menuitem);

	// menuitem == NULL: The locales_list in profile is empty. Only VTE_CJK_WIDTH is setted.
	// if (menuitem)
	//	g_debug("Add page by %p to locale %s", win_data->current_vte, gtk_widget_get_name(menuitem);

	// locale should not be free()
	const gchar *locale = NULL;
	gchar **locale_split = NULL;
	gchar *encoding = NULL;
	if (g_ascii_strcasecmp(gtk_widget_get_name(local_menuitem), "UTF-8"))
	{
		locale = gtk_widget_get_name(local_menuitem);
		if ((strcmp(locale, "C")==0) || (strcmp(locale, "POSIX")==0))
			encoding = "ANSI_X3.4-1968";
		else
		{
			locale_split = g_strsplit(locale, ".", -1);
			if (locale_split[0])
				encoding = locale_split[1];
		}
		// const gchar *LC_ALL = g_getenv ("LC_ALL");
		// set_env("LC_ALL", locale , TRUE);
		// g_debug("Set ENV LC_ALL to %s, now LC_ALL = %s!", locale, g_getenv ("LC_ALL"));
		// encoding = g_strdup(get_encoding_from_locale());
		// g_debug("Set the encoding to %s in new_tab_with_locale()", encoding);
		// set_env("LC_ALL", LC_ALL, TRUE);
		g_free(win_data->default_locale);
		win_data->default_locale = g_strdup(locale);
		// g_debug("Set win_data->default_locale = %s", win_data->default_locale);
	}
	
	add_page(win_data,
		 page_data,
		 NULL,
		 encoding,
		 (gchar *) locale,
		 NULL,
		 NULL,
		 FALSE,
		 get_VTE_CJK_WIDTH_str(VTE_CJK_WIDTH+1),
		 TRUE);
	
	g_strfreev (locale_split);
}

void select_font(GtkWidget *widget, GtkWidget *window)
{
#ifdef DETAIL
	g_debug("! Launch select_font() for window %p", window);
#endif
	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(menu_active_window), "Win_Data");
	struct Page *page_data = (struct Page *)g_object_get_data(G_OBJECT(win_data->current_vte), "Page_Data");
	
	GtkWidget *dialog = gtk_font_selection_dialog_new(_("Font Selection"));
	gtk_window_set_transient_for(GTK_WINDOW(dialog), GTK_WINDOW(window));
	// set the default font name in gtk_font_selection_dialog
	gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG(dialog), page_data->font_name);
	if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK)
	{
		// g_debug("Trying to change font name!\n");
		g_free(page_data->font_name);
		page_data->font_name = gtk_font_selection_dialog_get_font_name(GTK_FONT_SELECTION_DIALOG(dialog));
		set_vte_font(NULL, 7);
	}
	gtk_widget_destroy(dialog);
}

void refresh_profile_list (struct Window *win_data)
{
#ifdef DETAIL
	g_debug("! Launch refresh_profile_list() with win_data = %p", win_data);
#endif
	win_data->subitem_profile_list = gtk_menu_new ();
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (win_data->menuitem_profile_list), win_data->subitem_profile_list);
	create_profile_list(win_data->subitem_profile_list, GTK_STOCK_NEW, win_data);
}

void create_profile_list(GtkWidget *sub_menu, const gchar *stock_id, struct Window *win_data)
{
#ifdef DETAIL
	g_debug("! Launch create_profile_list() with sub_menu = %p, stock_id = %s, win_data = %p",
		sub_menu, stock_id, win_data);
#endif
	// LilyTerm Buildin Configures
	gchar *menu_label = g_strdup_printf(_("%s default"), PACKAGE_NAME);
	create_menu_item (0, sub_menu, menu_label, "", stock_id,
			  (GSourceFunc)load_user_settings_from_path, win_data);
	g_free(menu_label);

	// System Profile
	const char * const *system_dirs = g_get_system_config_dirs();
	gint j;
	for (j=0; system_dirs[j] != NULL; j++)
	{
		gchar *sys_profile = g_strdup_printf("%s/%s", system_dirs[j], SYS_PROFILE);
		create_menu_item (0, sub_menu, _("System default"), sys_profile, stock_id,
				  (GSourceFunc)load_user_settings_from_path, win_data);
		g_free(sys_profile);
	}

	// User's main profile
	// g_debug("profile_dir = %s", profile_dir);
	GDir *dir  = g_dir_open (profile_dir, 0, NULL);

	if (dir==NULL) goto FINISH;

	gboolean create_separator = TRUE;
	const gchar *entry = g_dir_read_name(dir);
	gchar *user_profile;
	if (entry==NULL) goto FINISH;
	do
	{
		if (create_separator)
		{
			add_separator_menu(sub_menu);
			create_separator = FALSE;
		}
		if (entry[0]=='.') continue;
		user_profile = g_strdup_printf("%s/%s", profile_dir, entry);
		create_menu_item (0, sub_menu, entry, user_profile, stock_id,
				  (GSourceFunc)load_user_settings_from_path, win_data);
		g_free(user_profile);
	}
	while ((entry = g_dir_read_name(dir)) != NULL);

FINISH:
	g_dir_close(dir);
}

GtkWidget *create_menu_item (gint type, GtkWidget *sub_menu, const gchar *label, const gchar *label_name, 
			     const gchar *stock_id, GtkFunction func, gpointer func_data)
{
#ifdef DETAIL
	g_debug("! Launch create_menu_item() with sub_menu = %p, label = %s, "
		"label_name = %s, stock_id = %s, func = %p, func_data = %p",
		sub_menu, label, label_name, stock_id, func, func_data);
#endif
	GtkWidget *menu_item = NULL;
	switch (type)
	{
		case 0:
			menu_item = gtk_image_menu_item_new_with_label(label);
			gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menu_item),
						      gtk_image_new_from_stock(stock_id, GTK_ICON_SIZE_MENU));
			break;
		case 1:
			menu_item = gtk_check_menu_item_new_with_label (label);
			break;
		default:
			return NULL;
	}
	// g_debug("set the name of menuitem to : %s", label_name);
	if (label_name) gtk_widget_set_name(menu_item, label_name);
	gtk_menu_shell_append(GTK_MENU_SHELL(sub_menu), menu_item);
	g_signal_connect(menu_item, "activate", G_CALLBACK(func), func_data);
	return menu_item;
}

GtkWidget *create_sub_item (GtkWidget *menu, gchar *label, const gchar *stock_id)
{
#ifdef DETAIL
	g_debug("! Launch create_sub_item() with menu = %p, label = %s, stock_id = %s", menu, label, stock_id);
#endif
	GtkWidget *menu_item = create_sub_item_subitem (label, stock_id);
	return create_sub_item_submenu (menu, menu_item);
}

GtkWidget *create_sub_item_subitem (gchar *label, const gchar *stock_id)
{
#ifdef DETAIL
	g_debug("! Launch create_sub_item_subitem() with label = %s, stock_id = %s", label, stock_id);
#endif
	GtkWidget *menu_item = gtk_image_menu_item_new_with_label(label);
	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menu_item),
				      gtk_image_new_from_stock(stock_id, GTK_ICON_SIZE_MENU));
	return menu_item;
}

GtkWidget *create_sub_item_submenu (GtkWidget *menu, GtkWidget *menu_item)
{
#ifdef DETAIL
	g_debug("! Launch create_sub_item() with menu = %p, menu_item = %p", menu, menu_item);
#endif
	GtkWidget *sub_menu = gtk_menu_new ();
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item), sub_menu);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	return sub_menu;
}

void add_separator_menu (GtkWidget *sub_menu)
{
#ifdef DETAIL
	g_debug("! Launch add_separator_menu() with sub_menu = %p", sub_menu);
#endif
	GtkWidget *menu_item = gtk_separator_menu_item_new();
	gtk_menu_shell_append (GTK_MENU_SHELL (sub_menu), menu_item);
}

long get_profile_dir_modtime()
{
#ifdef DETAIL
	g_debug("! Launch get_profile_dir_modtime()");
#endif
	struct stat buffer;

	if (stat(profile_dir, &buffer))
		return 0;
	return buffer.st_mtime;
}
